<?php 
/**
	* @command
	* @filename class-phpSyntaxCheck.php
	* @author REDINFO (lcy@redinfo.co.kr)
	* @date 2023-09-03
	* @comment PHP Syntax 체크 라이브러리
	* @wran __construct 를 제외하고, __를 붙인 함수는 사용하지 않는다.
**/
class phpSyntaxCheck{
	private $option = array(
		'rootDir'=> array(), // 디렉토리 탐방 ROOT DIR (없을 경우 현재 디렉토리 기준) (절대경로)
		'allowDirList'=> array(), // 허용 디렉토리 목록 (절대경로)
		'denyDirList'=> array(), // 제외 디렉토리 목록 (절대경로)
		'allowFileList'=> array(), // 허용가능 파일이름(확장자 포함), 없을 경우 전체
		'denyFileList'=> array(), // 제외할 파일이름(확장자 제외), 없을 경우 전체		
		'errlogDir'=>'./log', // 에러로그를 남길 디렉토리 (해당 디렉토리는 기본적으로 757 이상의 권한을 가져야함)

		'debug'=> false, // 디버깅시 출려할 경우 true, 그렇지 않을 경우 false

		// 파일체크 기능은 실행파일의 중요도에 따라 주의가 필요하다. (상황에 따라 검사시 DB 접속을 하지 않도록 파일상 예외처리를 하는게 중요)
		// Notice 와 같이 php -l 로 확인 불가능한 경고성 에러도 추출할 수 있다. 단 위 언급처럼 PHP를 직접 실행하기 때문에 주의가 필요하며 가공된 출력 데이터를 활용하기 때문에 발생되는 오류에 대해 100% 보장되지 않는다.
		'filechk'=> true, 
		'errkey'=> array('PHP Notice','PHP Fatal error','PHP Parse error','PHP Warning'), // filechk 가 false 시 체크하지 않는다.

	);

	// 응답용 데이터
	var $response = array();

	// 생성자
	function __construct($option = array()){

		$optionKeys = array_keys($this->option);
		foreach($optionKeys as $k=>$v){
			if( empty($option[$v]))	continue;
			if( !empty($option[$v]) && is_array($option[$v]) && count($option[$v]) > 0){ $this->option[$v] = $option[$v];  }
		}

		// 검사 root 디렉토리가 없을 경우 현재의 디렉토리를 설정
		if( count($this->option['rootDir']) < 1){
			$this->option['rootDir'][] = __DIR__;
		}

	}

	// 옵션설정
	function set($key = '', $value = array()){
		if( empty($key)){ return false; }
		if( !isset($this->option[$key])){ return false; }
		

		switch($key){
			// 배열개열
			case "rootDir":
			case "allowDirList":
			case "denyDirList":
			case "allowFileList":
			case "denyFileList":
			case "errkey":
				if(empty($value)){ return false; } 
				if( !is_array($value)){ $value = array($value); }
				$value = array_values(array_unique(array_filter($value)));

				if( in_array($key, array('rootDir','allowDirList','denyDirList')) > 0 && count($value) > 0 ){
					foreach($value as $k=>$v){
						$value[$k] = realpath($v);
					}
				}

			break;

			// bool 개열
			case "debug":
			case "filechk":
				if( !is_bool($value)) $value = false;
			break;

			default:
				return false;
			break;
		}

		$this->option[$key] = $value;
	}

	// 전체 실행
	function apply(){
		// 옵션설정에 따른 디렉토리 추출
		$this->response['dirList'] = $this->dirList();		
		if( empty($this->response['dirList']) || count($this->response['dirList']) < 1){ return false; }

		// 옵션설정에 파일목록을 추출 
		$this->response['fileList'] = $this->fileList($this->response['dirList']); 

		// 문법검사 
		$this->response['syntax'] = $this->syntax($this->response['fileList']);

		// 검사 가능 파일이 없을 경우 false
		if( empty($this->response['syntax'])) { $this->response = false; }

		return $this->response;
	}

	// PHP 문법 검사
	function syntax($fileList = array()){
		try{
			if( count($fileList) < 1){ throw new Exception("검사 가능한 파일이 없습니다."); }
			$result = array('file'=>array(),'errmsg'=> array(),'fmsg'=>array(),'errfile'=>array());
			foreach($fileList as $dir=>$ls){
				foreach($ls as $k=>$v){
					$filePath = $dir.'/'.$v;
					$result['file'][] = $filePath;

					if( is_file($filePath) !== true){ $result['fmsg'][] = '`'.$filePath.'` 파일이 존재하지 않음'; continue; }
					unset($output);
					exec("php -l ".$filePath." 2>&1",$output);
					$output = array_values(array_filter($output));
					foreach($output as $sk=>$sv){
						if( preg_match('/(No syntax errors detected)/i',$sv) > 0){ unset($output[$sk]); break; }
					}
					if( count($output) > 0){
						$result['errmsg'][] = $output[0];	
						$result['errfile'][$filePath] = '<strong style="color:red">'.$filePath.'</strong>';
					}
					else{
						$filechk = $this->filechk($filePath);
						if( $filechk !== true){  
							$result['errmsg'] = array_merge($result['errmsg'],$filechk); 
							$result['errfile'][$filePath] = '<strong style="color:red">'.$filePath.'</strong>';
						}
					}
				}
			}
		}
		catch(Exception $e){
			return $this->debug($e);
		}

		if( count($result['fmsg']) > 0){
			$result['fmsg'] = array_values(array_unique(array_filter($result['fmsg'])));
		}
		if( count($result['errmsg']) > 0){
			$result['errmsg'] = array_values(array_unique(array_filter($result['errmsg'])));
			foreach( $result['errmsg'] as $k=>$v){
				$result['errmsg'][$k] = str_replace(array_keys($result['errfile']), array_values($result['errfile']), $v);
			}
			
		}		

		return $result;
	}

	// PHP 파일체크
	function filechk($filePath = ''){
		if( !is_file($filePath) || $this->option['filechk'] === false ){ return true; }
		unset($output);
		exec("php -f ".$filePath." 2>&1",$output);
		$output = array_values(array_unique(array_filter($output)));
		if( count($output) < 1){ return true; }
		$filechk = array();
		$list = array();
		foreach($output as $k=>$v){		
			$ex = explode("<br>",trim(strip_tags($v)));
			$ex = array_values(array_unique(array_filter($ex)));

			if( count($ex) > 0){
				$list = array_merge($list,$ex);
			}
		}

		foreach($list as $k=>$v){
			if( preg_match('/(on line)/i',$v) && preg_match('/('.implode(")|(",$this->option['errkey']).')/i',$v) ){
				preg_match('/PHP(.*?)on line(.*?)[0-9]*$/',$v,$pq);
				$pq = array_filter($pq);
				if( count($pq = array_filter($pq)) > 0){ $filechk[] = $pq[0]; }
				continue;
			}
		}
		if( count($filechk) < 1){ return true;}
		return $filechk;
	}

	// 디렉토리 탐방
	function dirloop($dir, $dirList = array()){
		if( !is_dir($dir)){  return $dirList;} 
		$dl = glob($dir.'/*',GLOB_ONLYDIR); // 해당 디렉토리만
		foreach($dl as $k=>$v){
			$dsl = glob($v.'/*',GLOB_ONLYDIR); // 해당 디렉토리만
			$dirList[$v] = true;
			if( count($dsl) > 0){
				$dirList = $this->dirloop($v,$dirList);
			}
		}
		return $dirList;
	}

	// 디렉토리 초기화 (허용/비허용 디렉토리 체크)
	function dirInit($dirList = array()){
		if( count($dirList) < 1){ return false; }
		if( !is_array($this->option['allowDirList']) || count($this->option['allowDirList']) < 1 || $this->option['denyDirList'] < 1 ){ return $dirList; }

		$dirInit = array();
		$dirAllowTree = array();
		$dirDenyTree = array();
		if( count($this->option['allowDirList']) > 0){
			foreach($dirList as $k=>$v){
					if( in_array($k,$this->option['allowDirList']) > 0){
						$dirInit[$k] = true ;
						$dirAllowTree[$k] = true;
						continue;
					}
					$exDir = explode("/",$k);
					unset($exDir[count($exDir)-1]);
					$rootDir = implode("/",array_values($exDir));
					if(isset($dirAllowTree[$rootDir])){
						$dirInit[$k] = true ;
						$dirAllowTree[$k] = true;
					}
				}
		}

		if( count($this->option['denyDirList']) > 0){
			foreach($dirList as $k=>$v){
					if( in_array($k,$this->option['denyDirList']) > 0){
						unset($dirInit[$k]);
						$dirDenyTree[$k] = true;
						continue;
					}

					$exDir = explode("/",$k);
					unset($exDir[count($exDir)-1]);
					$rootDir = implode("/",array_values($exDir));
					if(isset($dirDenyTree[$rootDir])){
						unset($dirInit[$k]);
						$dirDenyTree[$k] = true;
					}
				}
		}

		return $dirInit;
	}

	// 디렉토리 전체 목록
	function dirList(){
		try{
			if( !is_array($this->option['rootDir']) || count($this->option['rootDir']) < 1){ throw new Exception("`rootDir` Empty"); }

			// root 디렉토리 목록 추출
			$dirList = array();
			foreach($this->option['rootDir'] as $k=>$v){
				if( !is_dir($v)){ continue; }
				if( preg_match("/[\/]$/",$v)){  $v = substr_replace($v,'',-1); }
				$dirList[$v] = true;	
				$dirList = $this->dirloop($v, $dirList);
			}
			if( count($dirList) < 1){ throw new Exception("수행가능한 ROOT 디렉토리 목록이 없습니다."); }	

			// 디렉토리 초기화
			$dirList = $this->dirInit($dirList);
			if( count($dirList) < 1){ throw new Exception("수행가능한 ROOT 디렉토리 목록이 없습니다."); }
		}
		catch(Exception $e){
			return $this->debug($e);
		}


		return $dirList;
	}

	// 파일목록 (허용/비허용 파일명 체크)
	function fileList($dirList = array()){
		try{
			if( count($dirList) < 1){ throw new Exception("디렉토리가 비었습니다."); }
			$fileList = array();
			foreach($dirList as $k=>$v){
				if( !isset($fileList[$k])){ $fileList[$k] = array(); }
				$fl = glob("{".$k."/*.php,".$k."/*.html}",GLOB_BRACE); // 해당 디렉토리에서 특정 파일만 
				if( count($fl) < 1){ continue; }
				foreach($fl as $sk=>$sv){

					// async 일경우 loop 되기 때문에 실행관련 파일들은 검사제외
					if( $sv == $_SERVER['SCRIPT_FILENAME']){ continue; } 
					if( $sv == __FILE__){ continue; } 

					$allowChk = true;
					$fn = basename($sv);
					if( count($this->option['allowFileList']) > 0){
						if( in_array($fn, $this->option['allowFileList']) < 1){ 
							$allowChk = false;
						}
					}					

					if( $allowChk === true && count($this->option['denyFileList']) > 0){
						if( in_array($fn, $this->option['denyFileList']) > 0){ 
							$allowChk = false;
						}
					}					

					if( $allowChk === true){
						$fileList[$k][] = $fn;
					}
				}
			}
		}
		catch(Exception $e){
			return $this->debug($e);
		}
		$fileList = array_filter($fileList);
		return $fileList;		
	}
	
	// 비동기 통신 
	function async($url = ''){
		if( !empty($_SERVER['SHLVL'])){  return false; } // filechk loop 방지
		if( empty($url)) return false;
		$parts=parse_url($url);  
		$host = $port = '';
		if($parts['scheme'] == 'http'){
			$host = $parts['host'];
			$port = !empty($parts['port']) ? $parts['port']:80; 
		}
		else if($parts['scheme'] == 'https')
		{  
			$host = 'ssl://'.$parts['host'];
			$port = !empty($parts['port']) ? $parts['port']:443;
		}  

		$fp = fsockopen( $host, $port, $errno, $errstr, 30);  

		if( !empty($parts['query'])){ $parts['path'] .= '?'.$parts['query']; }
		$out = "GET ".$parts['path']." HTTP/1.1\r\n";  
		$out.= "Host: ".$parts['host']."\r\n";  
		if(!empty($_SERVER["HTTP_USER_AGENT"])){$out.="User-Agent: ".$_SERVER["HTTP_USER_AGENT"]."\r\n";}
		$out .="Referer: ".$url."\r\n";
		if(!empty($_SERVER['HTTP_COOKIE'])){ $out.= "Cookie: ".$_SERVER['HTTP_COOKIE']."\r\n";  }
		$out.= "Content-Type: application/x-www-form-urlencoded\r\n";  
		$out.= "Content-Length: ".strlen($parts['query'])."\r\n\r\n";  
		$out.= "Connection: Close\r\n\r\n";  
		fwrite($fp, $out);
		fclose($fp); 
		return true;
	}

	// 변수보기 헬퍼함수
	function pr($var = ''){
		if( empty($var) ){ return false; }
		echo '<pre>';
		print_r($var);
		echo '</pre>';
	}

	// 에러로그
	function errlog($msg = '', $filename = ''){
		if( empty($filename)){ $filename = date('Ymd')."-error.log"; }
		if( empty($msg)){ return false; }
		if( is_dir($this->option['errlogDir']) == false){ return false; } // log 디렉토리는 757 이상의 권한 필요
		$fp = fopen($this->option['errlogDir']."/".$filename, 'a+');
		fwrite($fp, $msg);
		fclose($fp);        
		return true;
	}

	// 디버깅
	function debug($e = false, $return = false){
		$debug = array();
		$debug['rst'] = 'fail'; 

		if( is_object($e)){
			$debug['msg'] = $e->getMessage(); 
			$debug['code'] = $e->getCode(); 
			$debug['line'] = $e->getLine(); 
		}
		else{
			$debug = $e;
		}

		if( $this->option['debug'] === true){
			$this->pr($debug);
		}
		return $return;
	}

}